/*-
 * Copyright 2007 Jason Harmening
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 */

#ifndef __CX88_MPEG_IOCTL_H__
#define __CX88_MPEG_IOCTL_H__

#include "cx88_common_ioctl.h"

#define CX88_MPEG_DEFAULT_PACKET_SIZE 188
#define CX88_MPEG_DEFAULT_START_BYTE 0x47
#define CX88_MPEG_DEFAULT_START_BYTE_COUNT 1

/* Configure the driver */
typedef struct
{
   uint32_t in_packet_size; /* Packet size in bytes */
   uint32_t inout_num_packets; /* Total number of packets for which to allocate buffer space */
   uint32_t out_packets_per_buffer; /* Number of packets in each buffer (last buffer may be smaller) */
   uint8_t out_num_buffers; /* Number of buffers */
   
   /* Buffer configuration */
   union
   {
      struct
      {
         /* mmap() works at page granularity, so to acommodate page-rounding, 
          * applications will need to use out_mmap_size to determine the total number
          * of bytes to mmap(), and out_buffer_size to compute buffer offsets within
          * the mmap'ed region. */
         uint32_t out_buffer_size; 
         uint32_t out_mmap_size;
      } out_kernelbuf; /* Kernel-mode buffer allocation */
      void *in_userbuf; /* User-mode buffer allocation */
   } bufconfig;
   
   #define CX88_MPEG_BUF_KERNEL 0
   #define CX88_MPEG_BUF_USER 1
   uint8_t in_buftype; /* Select user- or kernel-mode buffers */

} cx88_mpeg_config;

/* Start cx88mpeg capture */
typedef struct
{  
   /* MPEG packet start byte--usually 0x47 */
   uint8_t in_start_byte;
   /* Number of start bytes that must be detected for packet to be considered valid--typically 1 */
   uint8_t in_start_byte_count;
   
   /* Frontend interface options */
   #define CX88_MPEG_PUNCTURED_CLOCK  (1 << 0)
   #define CX88_MPEG_SERIAL_INPUT     (1 << 1)
   #define CX88_MPEG_POS_POLARITY     (1 << 2)
   #define CX88_MPEG_LITTLE_ENDIAN    (1 << 3)
   #define CX88_MPEG_BYTEWIDE_SOP     (1 << 4)
   uint32_t in_options;
   
} cx88_mpeg_capture;

/* Wait for buffer to become available */
typedef struct
{
   uint32_t in_timeout_s; /* Maximum time to wait (0 -> no timeout) */

   uint8_t out_buffer_available; /* 1-based index of most recent available buffer. 0->no buffers available */
   
   /* Capture status */
   #define CX88_MPEG_CAP_NOERR     0
   #define CX88_MPEG_CAP_FATAL     (1 << 0)
   #define CX88_MPEG_CAP_TIMEOUT   (1 << 1)
   #define CX88_MPEG_BAD_PACKET    (1 << 2)
   #define CX88_MPEG_HW_OVERFLOW   (1 << 3)
   #define CX88_MPEG_BUS_ERROR     ((1 << 4) | CX88_MPEG_CAP_FATAL)
   #define CX88_MPEG_DMA_PROGRAM   ((1 << 5) | CX88_MPEG_CAP_FATAL)
   #define CX88_MPEG_WAIT_ABORTED  ((1 << 6) | CX88_MPEG_CAP_FATAL)
   uint8_t out_status;
  
   /* FIFO overflow and bad packet counts since last CX88_MPEG_WAIT */
   uint16_t out_num_overflows;
   uint16_t out_num_bad_packets;
   
   /* Packet count since start of transfer */
   uint64_t out_packet_count;
   
} cx88_mpeg_wait;


#define CX88_MPEG_CONFIG      _IOWR('M', 0, cx88_mpeg_config)
#define CX88_MPEG_CAPTURE     _IOW('M', 1, cx88_mpeg_capture)
#define CX88_MPEG_STOP        _IO('M', 2)
#define CX88_MPEG_WAIT        _IOWR('M', 3, cx88_mpeg_wait)

#endif
